import os

import twitter

class Client():

    twitter = None

    def __init__(self, name):
        # load Twitter app consumer details
        # TODO: move this to ini file
        consumer_file = os.path.expanduser("~/.mimicbot/%s/consumer" % name)
        consumer_key, consumer_secret = twitter.read_token_file(consumer_file)

        # load authentication
        auth_file = os.path.expanduser("~/.mimicbot/%s/auth" % name)
        if not os.path.exists(auth_file):
            # none exist, so get authentication details
            twitter.oauth_dance(
                "mimicbot", consumer_key, consumer_secret, auth_file)

        # authenticate
        oauth_token, oauth_secret = twitter.read_token_file(auth_file)
        self.twitter = twitter.Twitter(auth=twitter.OAuth(
            oauth_token, oauth_secret, consumer_key, consumer_secret))

    def get_latest_tweets(self, username):
        tweets = self.twitter.statuses.user_timeline(
            screen_name=username, count=200)
        return tweets

    def post(self, text):
        self.twitter.statuses.update(status=text)



import os
import random
import re
import csv
import html

import dateutil
from dateutil import parser

from MarkovText import MarkovText

# TODO: don't use them like this
from whoosh import fields, index, qparser, analysis, sorting
import whoosh.query

import click

import pprint
pp = pprint.PrettyPrinter(indent=4, depth=9)

class MarkovGenerator:

    dir = None

    index = None

    chain = None

    context = None

    def __init__(self, dir):
        self.dir = dir

    def get_index(self):
        stem_ana = analysis.StemmingAnalyzer()
        schema = fields.Schema(
            id=fields.ID(unique=True),
            datetime=fields.DATETIME(sortable=True),
            reply=fields.BOOLEAN,
            retweet=fields.BOOLEAN,
            text=fields.TEXT(analyzer=stem_ana, stored=True)
        )
        index_dir = os.path.join(self.dir, "index")
        if os.path.exists(index_dir):
            self.index = index.open_dir(index_dir)
        else:
            os.mkdir(index_dir)
            self.index = index.create_in(index_dir, schema)

    def search(self, query):
        self.get_index()

        import pprint
        pp = pprint.PrettyPrinter(indent=4, depth=9)

        with self.index.searcher() as searcher:
            # improve relevance! form query from keywords
            keywords = searcher.key_terms_from_text("text", query, numterms=20)
            keyword_query = " ".join(
                [keyword for keyword, score in keywords])

            print("keyword query: %s" % keyword_query)

            parser = qparser.QueryParser(
                "text", self.index.schema, group=qparser.OrGroup)
            query = parser.parse(keyword_query)

            restrict_retweets = whoosh.query.Term("retweet", True)

            results = searcher.search(query, mask=restrict_retweets, limit=100)
            for result in results:
                yield result["text"]

    def get_context(self, number=5):
        if self.context:
            return self.context

        self.get_index()

        from whoosh.qparser.dateparse import DateParserPlugin

        datetimes = sorting.FieldFacet("datetime", reverse=True)
        parser = qparser.QueryParser("text", self.index.schema)
        query = parser.parse("*")
        searcher = self.index.searcher()

        restrict_replies = whoosh.query.Term("reply", True)

        results = searcher.search(
            query, mask=restrict_replies, sortedby=datetimes, limit=number)

        click.secho("Adding to context...", fg="green")

        context = set()

        for result in results:
            text = result["text"]

            click.secho(text)

            import re

            # don't want these influence context
            # drop usernames
            text = re.sub(r"@[^ ]+", " ", text)
            # drop links
            text = re.sub(r"http(s?)://[^ ]+", " ", text)
            # drop cut-off text from the end of manual rts
            text = re.sub(r"[^ ]+…$", " ", text)

            context.update(text.split(" "))

        # split and discard empty strings
        context = " ".join(filter(None, context))

        click.secho("Processed context:", fg="green")
        click.secho(context)

        self.context = context

        return self.context

    def read_csv(self, filename):
        with open(filename, "r") as csv_file:
            reader = csv.reader(csv_file, delimiter=",", quotechar="\"")
            for row in reader:
                if row[0] == "tweet_id":
                    # skip header row
                    continue
                id = row[0]
                datetime = dateutil.parser.parse(row[3])
                reply = True if row[1] else False
                retweet = True if row[6] else False
                text = html.unescape(row[5])
                # turn newlines into something a markov chain is aware of
                text = re.sub(r"\n", " %NEWLINE% ", text)
                yield id, datetime, reply, retweet, text

    def train_csv(self, filename):
        self.get_index()
        writer = self.index.writer()
        for id, datetime, reply, retweet, text in self.read_csv(filename):
            print("writing doc: %s, %s, %s, %s, %s" % (id, datetime, reply, retweet, text))
            writer.update_document(id=id, datetime=datetime, reply=reply, retweet=retweet, text=text)
        writer.commit()
        doc_count = self.index.doc_count()
        print("doc count: %s" % doc_count)

    def train_latest_tweets(self, tweets):
        # TODO: fix duplication with train_csv method
        self.get_index()
        writer = self.index.writer()
        for tweet in tweets:
            id = tweet["id_str"]
            datetime = dateutil.parser.parse(tweet["created_at"])
            reply = True if tweet["in_reply_to_status_id"] else False
            retweet = True if tweet["retweeted"] else False
            text = html.unescape(tweet["text"])
            # turn newlines into something a markov chain is aware of
            text = re.sub(r"\n", " %NEWLINE% ", text)
            print("writing doc: %s, %s, %s, %s, %s" % (id, datetime, reply, retweet, text))
            writer.update_document(id=id, datetime=datetime, reply=reply, retweet=retweet, text=text)
        writer.commit()
        doc_count = self.index.doc_count()
        print("doc count: %s" % doc_count)

    def run(self, use_context, manual_context):
        # TODO needs to progressively search if we're not turning up enough
        # results
        context = ""
        if use_context:
            click.secho("Searching for context tweets...", fg="green")
            context = self.get_context()
        if manual_context:
            context = manual_context
        if not context:
            # TODO: better way of doing this
            context = "*"
        click.secho("Getting tweets for markov chain...", fg="green")
        results = list(self.search(context))
        # BIG TODO: EXPAND CONTEXT SEARCH SO WE AIM FOR X RESULTS
        # BIG TODO: (let's say 100)
        # BIG TODO: add this in ini file
        # BIG TODO: do this by expanding number of keywords generated from
        # BIG TODO context, until result size is good enough
        click.secho("Got %s tweets!" % len(results), fg="green")
        self.chain = MarkovText.Markov()
        for result in results:
            self.chain.add_to_dict(result)
        sentence_count = random.randint(1, 6)
        output = self.chain.create_sentences(sentence_count)
        output = re.sub(r"\s?%NEWLINE%\s?", "\n", output)
        return output




import os
import sys
import random
import time
import shutil

import click

import mimicbot
from mimicbot import twitter

@click.group()
def cli():
    pass

@cli.command()
@click.argument("bot_name")
def create(bot_name):
    "Create a bot."
    home = os.path.expanduser("~")
    bot_dir = os.path.join(home, ".mimicbot", bot_name)
    if not os.path.isdir(bot_dir):
        click.secho("Creating bot...", fg="green")
        os.makedirs(bot_dir)
        click.secho("Done!", fg="green")
    else:
        click.secho("Bot exists!", fg="red")

@cli.command()
def list():
    "List bots."
    home = os.path.expanduser("~")
    bots_dir = os.path.join(home, ".mimicbot")
    for filename in os.listdir(bots_dir):
        if os.path.isdir(os.path.join(bots_dir, filename)):
            click.echo(filename)

@cli.command()
@click.argument("bot_name")
def delete(bot_name):
    "Delete a bot."
    home = os.path.expanduser("~")
    bot_dir = os.path.join(home, ".mimicbot", bot_name)
    if os.path.isdir(bot_dir):
        click.secho("Deleting bot...", fg="green")
        shutil.rmtree(bot_dir)
        click.secho("Done!", fg="green")
    else:
        click.secho("Not a bot!", fg="red")

@cli.command()
@click.argument("bot_name")
def auth(bot_name):
    "Authenticate a bot against Twitter."
    # TODO: before doing this, wipe the existing auth file to force re-auth
    client = twitter.Client(bot_name)
    click.echo("auth")

@cli.command()
@click.argument("bot_name")
def reset():
    "Reset a bot's training."
    click.echo("reset")

@cli.command()
@click.option("--latest-tweets", metavar="USERNAME",
    help="Train from a user's latest tweets.")
@click.option("--csv-archive", metavar="FILENAME",
    help="Train from Twitter CSV tweet archive.")
@click.argument("bot_name")
def train(bot_name, latest_tweets, csv_archive):
    "Train a bot from source material."
    bot = mimicbot.Bot(bot_name)
    if latest_tweets:
        client = twitter.Client(bot_name)
        tweets = client.get_latest_tweets(latest_tweets)
        bot.generator.train_latest_tweets(tweets)
    if csv_archive:
        bot.generator.train_csv(csv_archive)
    # TODO error out if both are missing
    # TODO add option for just reading lines from a file. for when someone is
    # running a bot based off of someone else's account. given that our "latest
    # tweets" only grabs the last 200 each time

@cli.command()
# TODO make username a config ini setting
# TODO add help docs saying that "not authorized" can be thrown if the training
# TODO account is locked
@click.option("--train", metavar="USERNAME",
    help="Train the bot from recent tweets before running.")
@click.option("--use-context", is_flag=True,
    help="Set context from LENGTH recent tweets.")
@click.option("--manual-context",
    help="Set context as TEXT.")
@click.option("--random-exit", default=0,
    help="Randomly exit instead of running. 1/INTEGER chance of succeeding.")
@click.option("--random-delay", metavar="SECONDS",
    help="Run with a random delay.")
@click.option("--dry-run", is_flag=True,
    help="Dry run. Do not post.")
@click.argument("bot_name")
def run(bot_name, train, use_context, manual_context, random_exit, random_delay, dry_run):
    "Run a bot."
    if random_exit:
        roll = random.randint(1, random_exit)
        if roll != 1:
            click.secho("Randomly exiting!", fg="green")
            return
    bot = mimicbot.Bot(bot_name)
    if train:
        # TODO: remove duplication from command above
        client = twitter.Client(bot_name)
        tweets = client.get_latest_tweets(train)
        bot.generator.train_latest_tweets(tweets)
    click.secho("Getting text...", fg="green")
    # TODO move the twitter client stuff into the bot
    text = bot.get_text(use_context, manual_context)
    click.secho("Got text...", fg="green")
    click.echo("%s" % text)
    if random_delay:
        seconds = random.randint(1, random_delay)
        click.secho("Sleeping for %ss..." % seconds, fg="green")
        time.sleep(seconds)
    if dry_run:
        click.secho("Dry run. Exiting...", fg="green")
    else:
        click.secho("Posting...", fg="green")
        client = twitter.Client(bot_name)
        client.post(text)



import os

from mimicbot import filter, generate, process

__version__ = "0.0.9"

class Bot:

    dir = None

    generator = None

    filter = None

    processor = None

    def __init__(self, name):
        home = os.path.expanduser("~")
        self.dir = os.path.join(home, ".mimicbot", name)
        if not os.path.isdir(self.dir):
            raise Exception("bot dir does not exist")
        self.generator = generate.Generator(self.dir)
        self.filter = filter.Filter()
        self.processor = process.Processor()

    def _handle(self, text):
        self.filter.run(text)
        # print("filter passed")
        return self.processor.run(text)

    def get_text(self, use_context, manual_context):
        text = None
        for i in range(100000):
            text = self.generator.run(use_context, manual_context)
            # print("\ngenerated: %s" % text)
            try:
                return self._handle(text)
            except Exception as error:
                print("error: %s" % error)
                continue
        raise Exception("too many failed attempts to filter text")




import re

from pyparsing import nestedExpr, QuotedString

class Filter:

    def check_not_empty(self, text):
        if not text:
            raise Exception("empty string")

    def check_too_long(self, text):
        if len(text) > 140:
            raise Exception("too long")

    def check_contains_link(self, text):
        # update tests for only using ":"
        if re.findall("http(s?):", text):
            raise Exception("contains link")

    def check_contains_username(self, text):
        if re.findall("@\w+", text):
            raise Exception("contains username")

    def check_syntax(self, text):
        # delete contractions
        text = re.sub("\w('|’)\w", "", text)
        # valid syntax
        paired_exprs = \
            nestedExpr('(',')') | \
            nestedExpr('[',']') | \
            nestedExpr('{','}') | \
            QuotedString("\"") | \
            QuotedString("'") | \
            QuotedString("“", endQuoteChar="”") | \
            QuotedString("‘", endQuoteChar="’") | \
            QuotedString("*") \
        # strip out matched syntax
        stripped_line = paired_exprs.suppress().transformString(text)
        # if there are any quotes or parentheses left, they were not
        # properly nested
        if any(unwanted in stripped_line for unwanted in "()[]{}\"'“”‘’*"):
            raise Exception("bad syntax")

    def run(self, text):
        self.check_not_empty(text)
        self.check_too_long(text)
        self.check_contains_link(text)
        self.check_contains_username(text)
        self.check_syntax(text)

# import language_check

# tool = language_check.LanguageTool('en-GB')

# print(text)
# matches = tool.check(text)
# text = language_check.correct(text, matches)
